/*========================================================
=== Simple fractal landscape generator and viewer.
=== This isn't intended to be any general ultra-optimized
=== 3D engine. This is simple as what it does..
=== It can be generally optimized using the inline GTE
=== calls. It can be specifically optimized anyhow (the
=== ways of optimizations are infinite 8).
=== I'm generating static display lists. In a generic 3D
=== engine they should be dynamically allocated.
=== I'm using malloc() here because it works but remember
=== that it is buggy !! Write your own memory manager or
=== get the patch for the malloc() !!
=== This is my very first PSX program. I made it in one
=== Friday night freezing in the office. So, please, no
=== complains about lack of comments in the code !
=== This little header is signed 3/24/97 by
===
=== Davide Pasca
===  dpasca@ix.netcom.com - dpasca@val.net
===  http://www.netcom.com/~dpasca
===
=== P.S.   Tabs are 4 spaces over here...
=== P.P.S. "C" Rules !
*/

#include <sys/types.h>
#include <libetc.h>
#include <libgte.h>
#include <libgpu.h>
#include <rand.h>
#include <malloc.h>

#define OTLEN	12
#define OTSIZE	(1<<OTLEN)
#define NOT(_X_)	(!(_X_))
#define ABS(_X_)	((_X_)<0?-(_X_):(_X_))
#define CLAMP(A,MI,MA)	{if ((A)<(MI)) (A) = (MI); else if ((A)>(MA)) (A)=(MA);}

MATRIX	_identity = { ONE,0,0,
					  0,ONE,0,
					  0,0,ONE,
					  0,0,0 };

#define ALLOCWS(_D_,_S_)	{ _D_ = (void *)(_curdisplayp->wspace + _curdisplayp->wspacepos);\
							  _curdisplayp->wspacepos += _S_; }

/*============================================*/
typedef struct {		
	DRAWENV		draw;			/* drawing environment */
	DISPENV		disp;			/* display environment */
	u_long		ot[OTSIZE];		/* ordering table */
	char		wspace[1024*200];
	int			wspacepos;
} display_t;

/*============================================*/
static display_t	_display[2], *_curdisplayp;
static int			_curdisplayi;
static int			_hsyncs_height=1;

/*============================================*/
static void init_psx(void)
{
	ResetCallback();
	PadInit(0);         /* initialize PAD */
	InitGeom();			/* initialize geometry subsystem */
}

/*============================================*/
static void init_graph( int w, int h, int d )
{
	ResetGraph(0);		/* reset graphic subsystem (0:cold,1:warm) */
	SetGraphDebug(1);	/* set debug mode (0:off, 1:monitor, 2:dump) */	

	SetDefDrawEnv(&_display[0].draw, 0, 0, w, h);
	SetDefDrawEnv(&_display[1].draw, 0, h, w, h);
	SetDefDispEnv(&_display[0].disp, 0, h, w, h);
	SetDefDispEnv(&_display[1].disp, 0, 0, w, h);

	SetDispMask(1);		/* enable to display (0:inhibit, 1:enable) */

	_curdisplayi = 0;
	_curdisplayp = &_display[0];
	_display[0].wspacepos = _display[1].wspacepos = 0;

	_display[0].draw.isbg = _display[1].draw.isbg = 1;

	SetGeomOffset( w/2, h/2 );
	SetGeomScreen( d );

	PutDispEnv( &_curdisplayp->disp ); /* update display environment */

	if ( h == 240 )
		_hsyncs_height = 252;
	else
		_hsyncs_height = 504;
}

/*============================================*/
static void clear_screen( u_char r, u_char g, u_char b )
{
	_curdisplayp->draw.isbg = 1;
	setRGB0( &_curdisplayp->draw, r, g, b );
}

/*============================================*/
static long swap_buffers(void)
{
long	nlines;

	DrawSync(0);	/* wait for end of drawing */
	nlines = VSync(1);
	VSync(0);

	PutDrawEnv( &_curdisplayp->draw ); /* update drawing environment */
	PutDispEnv( &_curdisplayp->disp ); /* update display environment */
	/*DumpOTag(cdb->ot);	/* for debug */
	DrawOTag( _curdisplayp->ot );	/* draw */

	_curdisplayi ^= 1;
	_curdisplayp = &_display[ _curdisplayi ];
	_curdisplayp->wspacepos = 0;
	_curdisplayp->draw.isbg = 0;

	/* Use ClearOTagR() instead ! It's much faster ! */
	ClearOTag( _curdisplayp->ot, OTSIZE );	/* clear ordering table */

	return nlines;
}

/*============================================*/
/*============================================*/
/* This "plasma" thing is nothing but a fractal
   noise generator.
*/
#define RANDOM(_X_)	(rand()%(_X_))
#define PLASMA_PIX(_X_,_Y_)	_plasma_mapp[ (_Y_)*_plasma_mapwd + (_X_) ]
static short	*_plasma_mapp;
static int		_plasma_max, _plasma_mapwd, _plasma_noise;

/*============================================*/
static short _plasma_randpix( int x, int y, int x1, int y1, int x2, int y2 )
{
short	col;
int		dx,dy;

	dx = x - x1;
	dy = y - y1;
	col = (RANDOM(_plasma_max) - _plasma_max/2) *
		  (ABS(dx) + ABS(dy)) / _plasma_noise;
	col += (int)PLASMA_PIX(x1,y1) + PLASMA_PIX(x2,y2) >> 1;

	if ( col < 1 ) col = 1;
	else
	if ( col >= _plasma_max ) col = _plasma_max - 1;

	return PLASMA_PIX( x, y ) = col;
}

/*============================================*/
static void _plasma_do( int x1, int y1, int x2, int y2 )
{
int		x, y;
short	s, p;

	if (x2-x1 <= 1 && y2-y1 <= 1)
		return;

	x = x1 + x2 >> 1;
	y = y1 + y2 >> 1;

	if NOT( p = PLASMA_PIX( x, y1 ) )
		p = _plasma_randpix(x,y1,x1,y1,x2,y1);
	s = p;

	if NOT( p = PLASMA_PIX( x2, y ) )
		p = _plasma_randpix(x2,y,x2,y1,x2,y2);
	s += p;

	if NOT( p = PLASMA_PIX( x, y2 ) )
		p = _plasma_randpix(x,y2,x1,y2,x2,y2);
	s += p;

	if NOT( p = PLASMA_PIX( x1, y ) )
		p = _plasma_randpix(x1,y,x1,y1,x1,y2);
	s += p;

	if NOT( PLASMA_PIX( x, y ) )
		PLASMA_PIX( x, y ) = s >> 2;

	_plasma_do(x1,y1,x,y);
	_plasma_do(x,y1,x2,y);
	_plasma_do(x1,y,x,y2);
	_plasma_do(x,y,x2,y2);
}

/*============================================*/
static void plasma_make( short *mapp, int w,
						 int x1, int y1, int x2, int y2,
						 int noise, int seed, int max )
{
int		av, m1, m2;
short	*p;

	srand( seed );
	_plasma_noise = (noise ? w/noise : w);
	_plasma_max = max;
	_plasma_mapp = mapp;
	_plasma_mapwd = w;

	av = RANDOM(max);
	m1 = max/3;
	m2 = max/6;
	if ( !*(p = mapp + y1*w + x1) )	*p = av + RANDOM(m1)-m2;
	if ( !*(p = mapp + y1*w + x2) )	*p = av + RANDOM(m1)-m2;
	if ( !*(p = mapp + y2*w + x2) )	*p = av + RANDOM(m1)-m2;
	if ( !*(p = mapp + y2*w + x1) )	*p = av + RANDOM(m1)-m2;

	_plasma_do( x1, y1, x2, y2 );
}

/*============================================*/
#define MAP_SIZE	50
static short		_map[MAP_SIZE * MAP_SIZE];
static CVECTOR		_colormap[MAP_SIZE * MAP_SIZE];
static SVECTOR		_verts[ MAP_SIZE * MAP_SIZE ];
static DVECTOR		_screen[ MAP_SIZE * MAP_SIZE ];
static u_short		_pee[ MAP_SIZE * MAP_SIZE ];
static u_short		_trazeta[ MAP_SIZE * MAP_SIZE ];
static u_short		_traflags[ MAP_SIZE * MAP_SIZE ];

#define MAX_HEIGHT	1000
#define SEA_LEVEL	350

/*============================================*/
static void make_3d_map( SVECTOR *vertsp, const short *mapp, int maxh, int sealev )
{
int		i, j;
int		y, dy;
int		x, dx;
short	aftersealev;

	aftersealev = sealev * 150 / 100;

	y = -MAP_SIZE/2 * 100;
	dy = dx = 100;
	for(i=0; i < MAP_SIZE; ++i, y += dy)
	{
		x = -MAP_SIZE/2 * 100;
		for(j=0; j < MAP_SIZE; ++j, x += dx)
		{
			vertsp->vx = x;				/* x,z grid.. easy */
			vertsp->vz = y;

			if ( *mapp < sealev )		/* everything at and below the sea level */
				vertsp->vy = -sealev;	/* is set to the sea level */
			else
			if ( *mapp < aftersealev )	/* some custom scaling here */
				vertsp->vy = -(sealev + (*mapp - sealev) * 40 / 100);
			else
				vertsp->vy = -( (*mapp - sealev) * 120 / 100 + sealev);

			++mapp;
			++vertsp;
		}
	}
}

/*============================================*/
static void make_color_map( CVECTOR *colmapp, const short *mapp, int *maxhp, int *sealevp )
{
int		i, h;
int		sealev, sandlev, greelev, midlev, snowlev;
int		dsea, dsand, dgree, dmid, dsnow;
int		maxh = -10000000;

	for (i=MAP_SIZE*MAP_SIZE-1; i >= 0; --i)
		if ( mapp[ i ] > maxh )
			maxh = mapp[ i ];

	snowlev = maxh;
	midlev  = 90 * maxh / 100;
	greelev = 62 * maxh / 100;
	sandlev = 55 * maxh / 100;
	sealev  = 45 * maxh / 100;

	dsea = sealev - 0;
	dsand = sandlev - sealev;	if ( dsand == 0 )	dsand = 1;
	dgree = greelev - sandlev;	if ( dgree == 0 )	dgree = 1;
	dmid = midlev - greelev;	if ( dmid == 0 )	dmid = 1;
	dsnow = snowlev - midlev;	if ( dsnow == 0 )	dsnow = 1;
	for (i=MAP_SIZE*MAP_SIZE; i; --i, ++colmapp)
	{
		h = *mapp++;

		if ( h <= sealev )
		{
			colmapp->r = 0;
			colmapp->g = 20 - 20 * (sealev - h) / dsea;
			colmapp->b = (210 - 210 * (sealev - h) / dsea) + 30;
		}
		else
		if ( h <= sandlev )
		{
			colmapp->r = 30 * -(h - sandlev) / dsand + 80;
			colmapp->g = 30 * -(h - sandlev) / dsand + 65;
			colmapp->b = 0;
		}
		else
		if ( h <= greelev )
		{
			colmapp->r = 8 * (greelev - h) / dgree;
			colmapp->g = 100 * -(h - greelev) / dgree + 100;
			colmapp->b = 0;
		}
		else
		if ( h <= midlev )
		{
			colmapp->r = (30 - 30 * (midlev - h) / dmid) + 55;
			colmapp->g = (30 - 30 * (midlev - h) / dmid) + 40;
			colmapp->b = 0;
		}
		else
		{
			colmapp->r = (55 - 55 * (snowlev - h) / dsnow) + 180;
			colmapp->g = (55 - 55 * (snowlev - h) / dsnow) + 180;
			colmapp->b = (55 - 55 * (snowlev - h) / dsnow) + 185;
		}
	}

	*maxhp = maxh;
	*sealevp = sealev;
}

/*============================================*/
static void calc_normal( const SVECTOR *v0, const SVECTOR *v1, const SVECTOR *v2, SVECTOR *np )
{
SVECTOR	d01, d12;
VECTOR	n;

	d01.vx = v0->vx - v1->vx;
	d01.vy = v0->vy - v1->vy;
	d01.vz = v0->vz - v1->vz;
	d12.vx = v1->vx - v2->vx;
	d12.vy = v1->vy - v2->vy;
	d12.vz = v1->vz - v2->vz;

	n.vx = d01.vy * d12.vz - d12.vy * d01.vz;
	n.vy = d01.vz * d12.vx - d12.vz * d01.vx;
	n.vz = d01.vx * d12.vy - d12.vx * d01.vy;

	VectorNormalS( &n, np );
}

/*============================================*/
#define add_vector(A,B)	(A)->vx += (B)->vx,(A)->vy += (B)->vy,(A)->vz += (B)->vz;

/*============================================*/
static void prepare_polys( const SVECTOR *vertsp, const CVECTOR *cmapp, POLY_G3 *pg3, VECTOR *light, int ambient )
{
CVECTOR					c0, c1, c2, c3;
register const SVECTOR	*v0, *v1, *v2, *v3;
const SVECTOR			*runvertsp;
SVECTOR					*normalsp, *runnormp;
VECTOR					*vnormalsp, *runvnormp;
SVECTOR					*runsvnormp;
int						i;
register int			j;

	normalsp = (SVECTOR *)malloc( MAP_SIZE*MAP_SIZE * sizeof(SVECTOR) * 2 );
	runnormp = normalsp;
	runvertsp = vertsp;
	for(i=0; i < MAP_SIZE-1; ++i, ++vertsp)
		for(j=0; j < MAP_SIZE-1; ++j)
		{
			v0 = vertsp;
			v1 = vertsp + 1;
			v2 = vertsp + 1 + MAP_SIZE;
			v3 = vertsp + MAP_SIZE;
			calc_normal( v0, v1, v2, runnormp++ );
			calc_normal( v2, v3, v0, runnormp++ );

			++vertsp;
		}

	vnormalsp = (VECTOR *)malloc( MAP_SIZE*MAP_SIZE * sizeof(VECTOR) );
	memset( vnormalsp, 0, MAP_SIZE*MAP_SIZE * sizeof(VECTOR) );
	runvnormp = vnormalsp;
	runnormp = normalsp;
	runvertsp = vertsp;
	for(i=0; i < MAP_SIZE-1; ++i, ++runvnormp)
		for(j=0; j < MAP_SIZE-1; ++j)
		{
		register VECTOR	*n0, *n1, *n2, *n3;
		
			n0 = runvnormp;
			n1 = runvnormp + 1;
			n2 = runvnormp + 1 + MAP_SIZE;
			n3 = runvnormp + MAP_SIZE;

			add_vector( n0, &runnormp[0] );
			add_vector( n1, &runnormp[0] );
			add_vector( n2, &runnormp[0] );

			add_vector( n2, &runnormp[1] );
			add_vector( n3, &runnormp[1] );
			add_vector( n0, &runnormp[1] );

			++runnormp;
			++runnormp;
			++runvnormp;
		}
	free( normalsp );

	runvnormp = vnormalsp;
	runsvnormp = (SVECTOR *)vnormalsp;
	for(i=0; i < MAP_SIZE-1; ++i)
	{
		for(j=0; j < MAP_SIZE-1; ++j)
		{
			VectorNormalS( runvnormp, runsvnormp );
			runsvnormp->pad = light->vx * runsvnormp->vx + 
							  light->vy * runsvnormp->vy +
							  light->vz * runsvnormp->vz >> 12;
			
			if ( runsvnormp->pad < 0 )
				runsvnormp->pad = ambient;

			++runsvnormp;
			++runvnormp;
		}
		++runsvnormp;
		++runvnormp;
	}

	runsvnormp = (SVECTOR *)vnormalsp;
	for(i=0; i < MAP_SIZE-1; ++i, ++cmapp, ++runsvnormp)
		for(j=0; j < MAP_SIZE-1; ++j)
		{
		register SVECTOR	*n0, *n1, *n2, *n3;
		long				t;

			n0 = runsvnormp;
			n1 = runsvnormp + 1;
			n2 = runsvnormp + 1 + MAP_SIZE;
			n3 = runsvnormp + MAP_SIZE;

			c0 = cmapp[0];
			c1 = cmapp[1];
			c2 = cmapp[1 + MAP_SIZE];
			c3 = cmapp[MAP_SIZE];

			t = (long)c0.r * n0->pad >> 12; CLAMP( t, 0, 255 ); c0.r = t;
			t = (long)c0.g * n0->pad >> 12; CLAMP( t, 0, 255 ); c0.g = t;
			t = (long)c0.b * n0->pad >> 12; CLAMP( t, 0, 255 ); c0.b = t;
			t = (long)c1.r * n1->pad >> 12; CLAMP( t, 0, 255 ); c1.r = t;
			t = (long)c1.g * n1->pad >> 12; CLAMP( t, 0, 255 ); c1.g = t;
			t = (long)c1.b * n1->pad >> 12; CLAMP( t, 0, 255 ); c1.b = t;
			t = (long)c2.r * n2->pad >> 12; CLAMP( t, 0, 255 ); c2.r = t;
			t = (long)c2.g * n2->pad >> 12; CLAMP( t, 0, 255 ); c2.g = t;
			t = (long)c2.b * n2->pad >> 12; CLAMP( t, 0, 255 ); c2.b = t;
			t = (long)c3.r * n3->pad >> 12; CLAMP( t, 0, 255 ); c3.r = t;
			t = (long)c3.g * n3->pad >> 12; CLAMP( t, 0, 255 ); c3.g = t;
			t = (long)c3.b * n3->pad >> 12; CLAMP( t, 0, 255 ); c3.b = t;

			*((CVECTOR *)&pg3->r0) = c0;
			*((CVECTOR *)&pg3->r1) = c1;
			*((CVECTOR *)&pg3->r2) = c2;
			setPolyG3( pg3 );
			++pg3;

			*((CVECTOR *)&pg3->r0) = c2;
			*((CVECTOR *)&pg3->r1) = c3;
			*((CVECTOR *)&pg3->r2) = c0;
			setPolyG3( pg3 );
			++pg3;

			++runsvnormp;
			++cmapp;
		}
	free( vnormalsp );
}

/*============================================*/
static void insert_polys( register const DVECTOR *dvertsp, register const short *zp )
{
register POLY_G3		*pg3;
register u_long			*ot;
char					*startp;
register const DVECTOR	*v0, *v1, *v2, *v3;
register int			z;
short					z02;
int						i;
register int			j;

	ot = _curdisplayp->ot;
	startp = _curdisplayp->wspace + _curdisplayp->wspacepos;
	pg3 = (POLY_G3 *)startp;
	for(i=0; i < MAP_SIZE-1; ++i)
	{
		for(j=0; j < MAP_SIZE-1; ++j)
		{
			z02 = zp[0] + zp[1 + MAP_SIZE];

			v0 = dvertsp;
			v1 = dvertsp + 1;
			v2 = dvertsp + 1 + MAP_SIZE;
			v3 = dvertsp + MAP_SIZE;

			z = zp[1] + z02 >> 4;
			if ( (u_short)z < OTSIZE-1 )	/* allow z == 0 */
			{
				*((DVECTOR *)&pg3->x0) = *v0;
				*((DVECTOR *)&pg3->x1) = *v1;
				*((DVECTOR *)&pg3->x2) = *v2;
				addPrim( ot + OTSIZE-1-z, pg3 ); /* -1 for z == 0 */
			}
			++pg3;

			z = zp[MAP_SIZE] + z02 >> 4;
			if ( (u_short)z < OTSIZE-1 )
			{
				*((DVECTOR *)&pg3->x0) = *v2;
				*((DVECTOR *)&pg3->x1) = *v3;
				*((DVECTOR *)&pg3->x2) = *v0;
				addPrim( ot + OTSIZE-1-z, pg3 );
			}
			++pg3;

			++zp;
			++dvertsp;
		}
		++zp;
		++dvertsp;
	}

	_curdisplayp->wspacepos += (char *)pg3 - startp;
}

/*============================================*/
static void recalc_world( short seed )
{
int	sealev, maxh;
static VECTOR	light = { -4096, -4096, -1024 };

	memset( _map, 0, sizeof(_map) );
	plasma_make( (short *)_map, MAP_SIZE,
				 0, 0, MAP_SIZE-1, MAP_SIZE-1,
				 2, seed, MAX_HEIGHT );

	make_color_map( _colormap, _map, &maxh, &sealev );
	make_3d_map( _verts, _map, maxh, sealev );

	DrawSync(0);
	VectorNormal( &light, &light );
	prepare_polys( _verts, _colormap, (POLY_G3 *)_display[0].wspace, &light, 20*ONE/100 );
	prepare_polys( _verts, _colormap, (POLY_G3 *)_display[1].wspace, &light, 20*ONE/100 );
}

/*============================================*/
static void animate( MATRIX *m, short aspect )
{
u_long			pad = PadRead(1);
static VECTOR	_ani_tra = { 0, SEA_LEVEL, MAP_SIZE*125 };
static SVECTOR	_ani_ang = { 10*ONE/360, 0, 0 };
static int		autoyaw_timeout=200, autoworld_timeout=200, cnt;

	/* zoom */
	if (pad & PADLup)	_ani_tra.vz += MAP_SIZE*2;
	if (pad & PADLdown)	_ani_tra.vz -= MAP_SIZE*2;
	CLAMP( _ani_tra.vz, MAP_SIZE*10, MAP_SIZE*1000 );
	
	/* yaw */
	if (pad & PADRleft)		_ani_ang.vy += 16;
	if (pad & PADRright)	_ani_ang.vy -= 16;

	/* pitch */
	if (pad & PADRup)	_ani_ang.vx += 16;
	if (pad & PADRdown)	_ani_ang.vx -= 16;

	/* roll */
	if (pad & PADL1)	_ani_ang.vz -= 16;
	if (pad & PADR1)	_ani_ang.vz += 16;

	if (pad & PADR2)
		recalc_world( cnt + _ani_tra.vz );

	if ( pad )
	{
		autoyaw_timeout = 200;
		autoworld_timeout = 200;
	}

	if ( autoyaw_timeout == 0 )
	{
		_ani_ang.vy += 16;
		if ( --autoworld_timeout == 0 )
		{
			recalc_world( cnt + _ani_tra.vz );
			autoworld_timeout = 200;
		}
	}
	else
		--autoyaw_timeout;


	++cnt;

	/* do rotation matrix and mutiply scale matrix
	   for the aspect ratio.
	*/
	{
	MATRIX	smat, rmat;

		smat = _identity;
		smat.m[0][0] *= aspect;

		RotMatrix( &_ani_ang, &rmat );
		
		MulMatrix0( &smat, &rmat, m );
		m->t[0] = _ani_tra.vx;
		m->t[1] = _ani_tra.vy;
		m->t[2] = _ani_tra.vz;
	}
}

/*============================================*/
void main(void)
{
short		ang;
int			cnt=0;

	init_psx();
	init_graph( 640, 240, 256 );

	FntLoad(960,256);
	SetDumpFnt(FntOpen(32,32,640,240,0,64));

	recalc_world( 513 );

	while (1)
	{
	MATRIX	mat;
	long	nlines;

		animate( &mat, 2 );
		SetRotMatrix( &mat );
		SetTransMatrix( &mat );

		RotTransPersN( _verts, _screen, _trazeta, _pee, _traflags, MAP_SIZE*MAP_SIZE );
		insert_polys( _screen, _trazeta );
		clear_screen( 0, 0, 0 );
		
		nlines = swap_buffers();
		FntPrint("fps = %d\n", 60 * _hsyncs_height / nlines );
		FntPrint("npolys = %d nverts = %d\n", MAP_SIZE*MAP_SIZE*2, MAP_SIZE*MAP_SIZE );
		FntPrint("poly/sec = %d\n", MAP_SIZE*MAP_SIZE*2 * 60 * _hsyncs_height / nlines );
		FntFlush(-1);

		/*printf( "%ld\n", cnt );*/
		++cnt;
	}
}
